import json
import torch
from rdkit.Chem import AllChem
from torch_geometric.data import Data
from .tokenizer import Iupac_tokenizer,Selfies_tokenizer
from.atom_rep  import mol_to_graph_data_obj_simple




class ipcMapper(object):
    def __init__(self,ipc_dir,max_len):
        self.max_len = max_len
        self.ipc_dir = ipc_dir
        self.training = True 
        self.json_dict = json.load(open(ipc_dir))
        self.tokenizer = Iupac_tokenizer()
        self.cls_token = self.tokenizer.convert_tokens_to_ids(['[CLS]'])[0]
        self.eos_token = self.tokenizer.convert_tokens_to_ids(['[EOS]'])[0]


    def __getitem__(self, id_):
        text = self.json_dict[id_]
            
        output = [self.get_single_ipc(text)]
        
        return output


    def get_single_ipc(self,text,max_len=None):
        output = {}
        txt_tokens = self.tokenizer.tokenize(text)
        txt_tokens = self.tokenizer.convert_tokens_to_ids(txt_tokens)
        
        txt_tokens =self.get_padded_tokens(txt_tokens,max_len)
        output['iupac_tokens'] = txt_tokens
        return output    
    
    def get_padded_tokens(self,txt_tokens,max_len=None):
        
        max_len = self.max_len if  max_len is None else max_len
        txt_tokens = txt_tokens[:max_len]


        txt_tokens = [self.cls_token] + txt_tokens + [self.eos_token]  


        txt_tokens = torch.tensor(txt_tokens, dtype=torch.long)

        output = torch.zeros(max_len + 2, dtype=torch.long)
        output[:len(txt_tokens)] = txt_tokens
        return output    
    
    def detokenize(self, ids):
        return  self.tokenizer.convert_ids_to_tokens(ids)
    


class sfsMapper(object):
    def __init__(self,sfs_dir,max_len):
        self.max_len = max_len
        self.sfs_dir = sfs_dir
        self.training = True 
        self.json_dict = json.load(open(sfs_dir))
        self.tokenizer = Selfies_tokenizer()
        self.sfs_vocab = self.tokenizer.vocab
        self.cls_token = self.tokenizer.convert_tokens_to_ids(['[CLS]'])[0]
        self.mask_token = self.tokenizer.convert_tokens_to_ids(['[MASK]'])[0]
        self.eos_token = self.tokenizer.convert_tokens_to_ids(['[EOS]'])[0]


    def __getitem__(self, id_):
        selfies = self.json_dict[id_]
            
        output = [self.get_single_ipc(selfies)]
        
        return output


    def get_single_ipc(self,text,max_len=None):
        output = {}
        txt_tokens = self.tokenizer.tokenize(text)
        txt_tokens = self.tokenizer.convert_tokens_to_ids(txt_tokens)
        
        txt_tokens =self.get_padded_tokens(txt_tokens,max_len)

        maksed_tokens,label = self.mask_tokens(txt_tokens)
        
        output['selfies_tokens'] = txt_tokens
        output['masked_tokens'] = maksed_tokens
        output['label'] = label
        return output    
    
    def mask_tokens(self,txt_tokens):
        sfs_tokens_masked = []
        label = [0 for _ in range(len(txt_tokens))]
        for i in range(len(txt_tokens)):
            token = txt_tokens[i]
            from random import random,randint
            if token != 0:
                if random() < 0.15:
                    label[i] = token
                    rate = random()
                    if rate < 0.8:
                        sfs_tokens_masked.append(self.mask_token)
                    elif rate < 0.9:
                        sfs_tokens_masked.append(randint(5,len(self.sfs_vocab)-1))
                    else:
                        sfs_tokens_masked.append(token)
                else:
                    sfs_tokens_masked.append(token)

            else:
                sfs_tokens_masked.append(token)
        return torch.tensor(sfs_tokens_masked),torch.tensor(label)
    
    def get_padded_tokens(self,txt_tokens,max_len=None):
        
        max_len = self.max_len if  max_len is None else max_len
        txt_tokens = txt_tokens[:max_len]


        txt_tokens = [self.cls_token] + txt_tokens + [self.eos_token]  


        txt_tokens = torch.tensor(txt_tokens, dtype=torch.long)

        output = torch.zeros(max_len + 2, dtype=torch.long)
        output[:len(txt_tokens)] = txt_tokens
        return output    
    
    def detokenize(self, ids):
        return  self.tokenizer.convert_ids_to_tokens(ids)
    


class mlgMapper(object):
    def __init__(self,smi_dir):
        self.mlg_dir = smi_dir
        self.training = True 
        self.json_dict = json.load(open(smi_dir))
        


    def __getitem__(self, id_):
        smi = self.json_dict[id_]
            
        output = self.get_single_mg(smi)
        
        return output


    def get_single_mg(self,smi,max_len=None):
        rdkit_mol = AllChem.MolFromSmiles(smi)
        if rdkit_mol != None:  # ignore invalid mol objects
            data = mol_to_graph_data_obj_simple(rdkit_mol)

        output = data

        return output    
    

    

from torch.utils.data import Dataset

class SIGDataset(Dataset):
    def __init__(self, ids_path, ipc_mapper, sfs_mapper, mlg_mapper, training):
        self.ipc_mapper = ipc_mapper
        self.sfs_mapper = sfs_mapper
        self.mlg_mapper = mlg_mapper
        self.ids = json.load(open(ids_path)) 
        self.idx = list(range(len(self.ids)))
        
    def __len__(self):
        return len(self.ids)

    def __getitem__(self, i):
        id_ = list(self.ids.keys())[i]
        smi = list(self.ids.values())[i]
        ipc_tokens = None 
        sfs_tokens = None
        mlg = None


        if AllChem.MolFromSmiles(smi):
            try:
                if self.ipc_mapper is not None:
                    ipc_tokens = self.ipc_mapper[id_]


                
                if self.sfs_mapper is not None:
                    sfs_tokens = self.sfs_mapper[id_]




                if self.mlg_mapper is not None:
                    mlg = self.mlg_mapper[id_]
            except:
                ipc_tokens = None 
                sfs_tokens = None
                mlg = None   



        return id_, sfs_tokens,ipc_tokens,mlg
    

from toolz.sandbox import unzip
from torch.utils.data import DataLoader

def SIG_collate(inputs):
    

    (id_, sfs_tokens,ipc_tokens,mlg) = map(list, unzip(inputs))


    if sfs_tokens[0] is not None:
        sfs_tokens = [ j  for i in sfs_tokens for j in i]
        sfs_tokens_collate = {}
        for k in sfs_tokens[0].keys():  
            sfs_tokens_collate[k] = torch.stack([i[k] for i in sfs_tokens]) 
    else:
        sfs_tokens_collate = None 
    

    if ipc_tokens[0] is not None:
        ipc_tokens = [ j  for i in ipc_tokens for j in i]
        ipc_tokens_collate = {}
        for k in ipc_tokens[0].keys():  #### bert tokens and clip tokens
            ipc_tokens_collate[k] = torch.stack([i[k] for i in ipc_tokens]) 
    else:
        ipc_tokens_collate = None 

    if mlg[0] is not None:        
        from torch_geometric.data import Batch
        mlg_ = Batch.from_data_list(mlg)
    else:
        mlg_ = None

        

    batch =   {'ids': id_,
             'sfs_tokens': sfs_tokens_collate,
             'ipc_tokens': ipc_tokens_collate,
             'mlg': mlg_}
    
    return batch



class FinetuneDataset(Dataset):
    def __init__(self, ids_path, smi_path ,ipc_mapper, sfs_mapper, mlg_mapper, testing):
        self.ipc_mapper = ipc_mapper
        self.sfs_mapper = sfs_mapper
        self.mlg_mapper = mlg_mapper
        self.smiles = json.load(open(smi_path))
        self.ids = json.load(open(ids_path)) 
        self.idx = list(range(len(self.ids)))
        
    def __len__(self):
        return len(self.ids)

    def __getitem__(self, i):
        id_ = list(self.ids.keys())[i]
        label = list(self.ids.values())[i]
        smi = list(self.smiles.values())[i]
        ipc_tokens = None 
        sfs_tokens = None
        mlg = None


   
        if self.ipc_mapper is not None:
            ipc_tokens = self.ipc_mapper[id_]


        
        if self.sfs_mapper is not None:
            sfs_tokens = self.sfs_mapper[id_]




        if self.mlg_mapper is not None:
            mlg = self.mlg_mapper[id_]



        return id_,smi, sfs_tokens,ipc_tokens,mlg,label
    
def Finetune_collate_C(inputs):
    (id_,smi,sfs_tokens,ipc_tokens,mlg,label) = map(list, unzip(inputs))

    label = torch.tensor(label, dtype=torch.long)
    if sfs_tokens[0] is not None:
        sfs_tokens = [ j  for i in sfs_tokens for j in i]
        sfs_tokens_collate = {}
        for k in sfs_tokens[0].keys():  
            sfs_tokens_collate[k] = torch.stack([i[k] for i in sfs_tokens]) 
    else:
        sfs_tokens_collate = None 
    

    if ipc_tokens[0] is not None:
        ipc_tokens = [ j  for i in ipc_tokens for j in i]
        ipc_tokens_collate = {}
        for k in ipc_tokens[0].keys():  #### bert tokens and clip tokens
            ipc_tokens_collate[k] = torch.stack([i[k] for i in ipc_tokens]) 
    else:
        ipc_tokens_collate = None 

    if mlg[0] is not None:        
        from torch_geometric.data import Batch
        mlg_ = Batch.from_data_list(mlg)

        

    batch =   {'ids': id_,
             'sfs_tokens': sfs_tokens_collate,
             'ipc_tokens': ipc_tokens_collate,
             'mlg': mlg_,
             'label': label}
    
    return batch

def Finetune_collate_R(inputs):
    (id_,smi,sfs_tokens,ipc_tokens,mlg,label) = map(list, unzip(inputs))

    label = torch.tensor(label, dtype=torch.float32)
    if sfs_tokens[0] is not None:
        sfs_tokens = [ j  for i in sfs_tokens for j in i]
        sfs_tokens_collate = {}
        for k in sfs_tokens[0].keys():  
            sfs_tokens_collate[k] = torch.stack([i[k] for i in sfs_tokens]) 
    else:
        sfs_tokens_collate = None 
    

    if ipc_tokens[0] is not None:
        ipc_tokens = [ j  for i in ipc_tokens for j in i]
        ipc_tokens_collate = {}
        for k in ipc_tokens[0].keys():  #### bert tokens and clip tokens
            ipc_tokens_collate[k] = torch.stack([i[k] for i in ipc_tokens]) 
    else:
        ipc_tokens_collate = None 

    if mlg[0] is not None:        
        from torch_geometric.data import Batch
        mlg_ = Batch.from_data_list(mlg)

        

    batch =   {'ids': id_,
             'sfs_tokens': sfs_tokens_collate,
             'ipc_tokens': ipc_tokens_collate,
             'mlg': mlg_,
             'label': label}
    
    return batch


import torch
import numpy as np
from rdkit import Chem
from torch_geometric.data import Data


# allowable node and edge features
allowable_features = {
    'possible_atomic_num_list' : list(range(1, 119)),
    'possible_formal_charge_list' : [-5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5],
    'possible_chirality_list' : [
        Chem.rdchem.ChiralType.CHI_UNSPECIFIED,
        Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CW,
        Chem.rdchem.ChiralType.CHI_TETRAHEDRAL_CCW,
        Chem.rdchem.ChiralType.CHI_OTHER
    ],
    'possible_hybridization_list' : [
        Chem.rdchem.HybridizationType.S,
        Chem.rdchem.HybridizationType.SP, Chem.rdchem.HybridizationType.SP2,
        Chem.rdchem.HybridizationType.SP3, Chem.rdchem.HybridizationType.SP3D,
        Chem.rdchem.HybridizationType.SP3D2, Chem.rdchem.HybridizationType.UNSPECIFIED
    ],
    'possible_numH_list' : [0, 1, 2, 3, 4, 5, 6, 7, 8],
    'possible_implicit_valence_list' : [0, 1, 2, 3, 4, 5, 6],
    'possible_degree_list' : [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
    'possible_bonds' : [
        Chem.rdchem.BondType.SINGLE,
        Chem.rdchem.BondType.DOUBLE,
        Chem.rdchem.BondType.TRIPLE,
        Chem.rdchem.BondType.AROMATIC
    ],
    'possible_bond_dirs' : [ # only for double bond stereo information
        Chem.rdchem.BondDir.NONE,
        Chem.rdchem.BondDir.ENDUPRIGHT,
        Chem.rdchem.BondDir.ENDDOWNRIGHT
    ]
}

def mol_to_graph_data_obj_simple(mol):
    """
    Converts rdkit mol object to graph Data object required by the pytorch
    geometric package. NB: Uses simplified atom and bond features, and represent
    as indices
    :param mol: rdkit mol object
    :return: graph data object with the attributes: x, edge_index, edge_attr
    """
    # atoms
    num_atom_features = 2   # atom type,  chirality tag
    atom_features_list = []
    for atom in mol.GetAtoms():
        atom_feature = [allowable_features['possible_atomic_num_list'].index(
            atom.GetAtomicNum())] + [allowable_features[
            'possible_chirality_list'].index(atom.GetChiralTag())]
        atom_features_list.append(atom_feature)
    x = torch.tensor(np.array(atom_features_list), dtype=torch.long)

    # bonds
    num_bond_features = 2   # bond type, bond direction
    if len(mol.GetBonds()) > 0: # mol has bonds
        edges_list = []
        edge_features_list = []
        for bond in mol.GetBonds():
            i = bond.GetBeginAtomIdx()
            j = bond.GetEndAtomIdx()
            edge_feature = [allowable_features['possible_bonds'].index(
                bond.GetBondType())] + [allowable_features[
                                            'possible_bond_dirs'].index(
                bond.GetBondDir())]
            edges_list.append((i, j))
            edge_features_list.append(edge_feature)
            edges_list.append((j, i))
            edge_features_list.append(edge_feature)

        # data.edge_index: Graph connectivity in COO format with shape [2, num_edges]
        edge_index = torch.tensor(np.array(edges_list).T, dtype=torch.long)

        # data.edge_attr: Edge feature matrix with shape [num_edges, num_edge_features]
        edge_attr = torch.tensor(np.array(edge_features_list),
                                 dtype=torch.long)
    else:   # mol has no bonds
        edge_index = torch.empty((2, 0), dtype=torch.long)
        edge_attr = torch.empty((0, num_bond_features), dtype=torch.long)

    data = Data(x=x, edge_index=edge_index, edge_attr=edge_attr)

    return data


import torch
from torch_geometric.nn import MessagePassing
from torch_geometric.utils import add_self_loops, degree, softmax
from torch_geometric.nn import global_add_pool, global_mean_pool, global_max_pool, GlobalAttention, Set2Set
import torch.nn.functional as F
from torch_scatter import scatter_add
from torch_geometric.nn.inits import glorot, zeros

# num_atom_type = 120 #including the extra mask tokens
# num_chirality_tag = 3

# num_bond_type = 6 #including aromatic and self-loop edge, and extra masked tokens
# num_bond_direction = 3 

class GINConv(MessagePassing):
    """
    Extension of GIN aggregation to incorporate edge information by concatenation.
    Args:
        emb_dim (int): dimensionality of embeddings for nodes and edges.
        embed_input (bool): whether to embed input or not. 
        
    See https://arxiv.org/abs/1810.00826
    """
    def __init__(self, emb_dim, aggr = "add"):
        super(GINConv, self).__init__()

        num_atom_type = 120 #including the extra mask tokens
        num_chirality_tag = 3

        num_bond_type = 6 #including aromatic and self-loop edge, and extra masked tokens
        num_bond_direction = 3 
        #multi-layer perceptron
        self.mlp = torch.nn.Sequential(torch.nn.Linear(emb_dim, 2*emb_dim), torch.nn.ReLU(), torch.nn.Linear(2*emb_dim, emb_dim))
        self.edge_embedding1 = torch.nn.Embedding(num_bond_type, emb_dim)
        self.edge_embedding2 = torch.nn.Embedding(num_bond_direction, emb_dim)

        torch.nn.init.xavier_uniform_(self.edge_embedding1.weight.data)
        torch.nn.init.xavier_uniform_(self.edge_embedding2.weight.data)
        self.aggr = aggr

    def forward(self, x, edge_index, edge_attr):
        #add self loops in the edge space
        edge_index = add_self_loops(edge_index, num_nodes = x.size(0))

        #add features corresponding to self-loop edges.
        self_loop_attr = torch.zeros(x.size(0), 2)
        self_loop_attr[:,0] = 4 #bond type for self-loop edge
        self_loop_attr = self_loop_attr.to(edge_attr.device).to(edge_attr.dtype)
        edge_attr = torch.cat((edge_attr, self_loop_attr), dim = 0)

        edge_embeddings = self.edge_embedding1(edge_attr[:,0]) + self.edge_embedding2(edge_attr[:,1])

        return self.propagate(edge_index[0], x=x, edge_attr=edge_embeddings)

    def message(self, x_j, edge_attr):
        return x_j + edge_attr

    def update(self, aggr_out):
        return self.mlp(aggr_out)

class GNN(torch.nn.Module):
    """
    
    Args:
        num_layer (int): the number of GNN layers
        emb_dim (int): dimensionality of embeddings
        JK (str): last, concat, max or sum.
        max_pool_layer (int): the layer from which we use max pool rather than add pool for neighbor aggregation
        drop_ratio (float): dropout rate
        gnn_type: gin, gcn, graphsage, gat
    Output:
        node representations
    """
    def __init__(self, num_layer, emb_dim, JK = "last", drop_ratio = 0, gnn_type = "gin"):
        super(GNN, self).__init__()
        self.num_layer = num_layer
        self.drop_ratio = drop_ratio
        self.JK = JK
        num_atom_type = 120 #including the extra mask tokens
        num_chirality_tag = 3


        if self.num_layer < 2:
            raise ValueError("Number of GNN layers must be greater than 1.")

        self.x_embedding1 = torch.nn.Embedding(num_atom_type, emb_dim)
        self.x_embedding2 = torch.nn.Embedding(num_chirality_tag, emb_dim)

        torch.nn.init.xavier_uniform_(self.x_embedding1.weight.data)
        torch.nn.init.xavier_uniform_(self.x_embedding2.weight.data)

        ###List of MLPs
        self.gnns = torch.nn.ModuleList()
        for layer in range(num_layer):
            self.gnns.append(GINConv(emb_dim, aggr = "add"))


        ###List of batchnorms
        self.batch_norms = torch.nn.ModuleList()
        for layer in range(num_layer):
            self.batch_norms.append(torch.nn.BatchNorm1d(emb_dim))

    def forward(self, *argv):
        if len(argv) == 3:
            x, edge_index, edge_attr = argv[0], argv[1], argv[2]
        elif len(argv) == 1:
            data = argv[0]
            x, edge_index, edge_attr = data.x, data.edge_index, data.edge_attr
        else:
            raise ValueError("unmatched number of arguments.")

        x = self.x_embedding1(x[:,0]) + self.x_embedding2(x[:,1])

        h_list = [x]
        for layer in range(self.num_layer):
            h = self.gnns[layer](h_list[layer], edge_index, edge_attr)
            h = self.batch_norms[layer](h)
            #h = F.dropout(F.relu(h), self.drop_ratio, training = self.training)
            if layer == self.num_layer - 1:
                #remove relu for the last layer
                h = F.dropout(h, self.drop_ratio, training = self.training)
            else:
                h = F.dropout(F.relu(h), self.drop_ratio, training = self.training)
            h_list.append(h)

        ### Different implementations of Jk-concat
        if self.JK == "concat":
            node_representation = torch.cat(h_list, dim = 1)
        elif self.JK == "last":
            node_representation = h_list[-1]
        elif self.JK == "max":
            h_list = [h.unsqueeze_(0) for h in h_list]
            node_representation = torch.max(torch.cat(h_list, dim = 0), dim = 0)[0]
        elif self.JK == "sum":
            h_list = [h.unsqueeze_(0) for h in h_list]
            node_representation = torch.sum(torch.cat(h_list, dim = 0), dim = 0)[0]

        return node_representation    
class GNN_graphpred(torch.nn.Module):
    """
    Extension of GIN to incorporate edge information by concatenation.
    Args:
        num_layer (int): the number of GNN layers
        emb_dim (int): dimensionality of embeddings
        num_tasks (int): number of tasks in multi-task learning scenario
        drop_ratio (float): dropout rate
        JK (str): last, concat, max or sum.
        graph_pooling (str): sum, mean, max, attention, set2set
        gnn_type: gin, gcn, graphsage, gat
        
    See https://arxiv.org/abs/1810.00826
    JK-net: https://arxiv.org/abs/1806.03536
    """
    def __init__(self,config=None):
        super(GNN_graphpred, self).__init__()

        self.num_layer = config.num_layer
        self.drop_ratio = config.drop_ratio
        self.JK = config.JK
        self.emb_dim = config.emb_dim
        self.num_tasks = config.num_tasks
        self.graph_pooling = config.graph_pooling
        self.gnn_type = config.gnn_type
        graph_pooling = config.graph_pooling

        if self.num_layer < 2:
            raise ValueError("Number of GNN layers must be greater than 1.")

        self.gnn = GNN(self.num_layer, self.emb_dim, self.JK, self.drop_ratio, gnn_type = self.gnn_type)

        #Different kind of graph pooling
        if graph_pooling == "sum":
            self.pool = global_add_pool
        elif graph_pooling == "mean":
            self.pool = global_mean_pool
        elif graph_pooling == "max":
            self.pool = global_max_pool
        elif graph_pooling == "attention":
            if self.JK == "concat":
                self.pool = GlobalAttention(gate_nn = torch.nn.Linear((self.num_layer + 1) * self.emb_dim, 1))
            else:
                self.pool = GlobalAttention(gate_nn = torch.nn.Linear(self.emb_dim, 1))
        elif graph_pooling[:-1] == "set2set":
            set2set_iter = int(graph_pooling[-1])
            if self.JK == "concat":
                self.pool = Set2Set((self.num_layer + 1) * self.emb_dim, set2set_iter)
            else:
                self.pool = Set2Set(self.emb_dim, set2set_iter)
        else:
            raise ValueError("Invalid graph pooling type.")

        #For graph-level binary classification
        if graph_pooling[:-1] == "set2set":
            self.mult = 2
        else:
            self.mult = 1
        
        if self.JK == "concat":
            self.graph_pred_linear = torch.nn.Linear(self.mult * (self.num_layer + 1) * self.emb_dim, self.num_tasks)
        else:
            self.graph_pred_linear = torch.nn.Linear(self.mult * self.emb_dim, self.num_tasks)

    def from_pretrained(self, model_file):
        #self.gnn = GNN(self.num_layer, self.emb_dim, JK = self.JK, drop_ratio = self.drop_ratio)
        self.gnn.load_state_dict(torch.load(model_file))

    def atom_feature_pad(self,input_,batch_size,max_len,index,embedding_size):
        output = torch.zeros(batch_size*max_len,embedding_size).to(input_.device)
        mask = torch.zeros(batch_size,max_len).to(input_.device)
        j= 0
        for i in range(batch_size):
            k = 0 
            while index[j]!=i:
                j+=1
            while j<len(index) and index[j]==i and k<max_len:
                output[i*max_len+k]=input_[j] 
                mask[i][k]=1
                k+=1
                j+=1
        # print(mask)
        
        return output.reshape(batch_size,max_len,embedding_size),mask

    def forward(self, *argv):
        if len(argv) == 4:
            x, edge_index, edge_attr, batch = argv[0], argv[1], argv[2], argv[3]
        elif len(argv) == 1:
            data = argv[0]
            x, edge_index, edge_attr, batch = data.x, data.edge_index, data.edge_attr, data.batch
        else:
            raise ValueError("unmatched number of arguments.")

        node_representation = self.gnn(x, edge_index, edge_attr)

        max_atom_num = 50

        node_feature,mask = self.atom_feature_pad(node_representation,batch[-1]+1,max_atom_num,batch,self.emb_dim)

        return self.pool(node_representation, batch),node_feature,mask


import torch
import pubchempy
import pandas as pd
from tqdm import tqdm
from concurrent.futures import ThreadPoolExecutor, as_completed
import time
from torch_geometric.data import InMemoryDataset
import selfies as sf

class PubChemDataset(InMemoryDataset):
    def __init__(self, path):
        super(PubChemDataset, self).__init__()
        self.data, self.slices = torch.load(path)
    
    def __getitem__(self, idx):
        return self.get(idx)

# 加载数据集
data = PubChemDataset('/home/hukun/code/Tmm-llama/data/PubChem324kV2/PubChem324kV2/pretrain.pt')

# 读取已经保存的 CSV 文件，获取已经爬取的 CID
def get_existing_cids(csv_file):
    try:
        df = pd.read_csv(csv_file)
        return set(df['CID'].dropna().astype(str))  # 返回已爬取数据的CID集合
    except FileNotFoundError:
        return set()  # 如果没有文件，则返回空集合

# 函数：获取iUPAC名称，添加超时处理
def get_iupac_name(smiles, timeout=60):
    try:
        compounds = pubchempy.get_compounds(smiles, namespace='smiles', timeout=timeout)
        if compounds:
            return compounds[0].iupac_name
        return None
    except Exception as e:
        print(f"Error fetching IUPAC name for {smiles}: {e}")
        return None

# 并行获取iUPAC名称
def fetch_data_for_index(i):
    sample = data[i]
    smiles = sample["smiles"]
    cid = sample["cid"]
    text = sample["text"]
    selfies_ = sf.encoder(smiles)
    # 获取 iupac name
    iupac_name = get_iupac_name(smiles)
    
    if iupac_name is not None:
        return [cid, smiles, text, iupac_name,selfies_, None]  # 返回数据
    else:
        return None  # 返回失败数据

# 打开一个 CSV 文件以便实时保存
csv_file = './data/pubchem324k/pubchem_data_2.csv'
df_columns = ["CID", "SMILES", "text", "iupac name","SELFIES"]

# 获取已经爬取过的 CID
existing_cids = get_existing_cids(csv_file)

# 使用 ThreadPoolExecutor 来并行处理数据
with ThreadPoolExecutor(max_workers=8) as executor:
    futures = []
    
    # 遍历数据集，为每个未爬取过的 CID 提交任务
    for i in range(0,len(data)):
        sample = data[i]
        cid = sample["cid"]
        if str(cid) not in existing_cids:
            futures.append(executor.submit(fetch_data_for_index, i))
    
    # 处理完成的任务
    for future in tqdm(as_completed(futures), desc="Processing", total=len(futures), unit="sample"):
        result = future.result()
        
        if result:
            cid, smiles, text, iupac_name, selfies_ = result
            # 保存有效数据
            temp_df = pd.DataFrame([[cid, smiles, text, iupac_name, selfies_]], columns=df_columns)
            temp_df.to_csv(csv_file, mode='a', header=False, index=False)

print("数据已成功保存为 CSV 文件！")




from .data import SIGDataset, SIG_collate, ipcMapper, sfsMapper, mlgMapper, FinetuneDataset, Finetune_collate_C, Finetune_collate_R
from .data4csv import Data4CSV, Data4CSV_collate
from torch.utils.data import DataLoader
import torch
from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.Chem.Scaffolds.MurckoScaffold import MurckoScaffoldSmiles
from torch.utils.data import Subset

def _generate_scaffold(smiles, include_chirality=False):
    mol = Chem.MolFromSmiles(smiles)
    scaffold = MurckoScaffoldSmiles(mol=mol, includeChirality=include_chirality)
    return scaffold

def generate_scaffolds(dataset, log_every_n=1000):
    scaffolds = {}
    data_len = len(dataset)
    print(data_len)

    print("About to generate scaffolds")
    for ind, _ in enumerate(dataset):
        smiles = dataset[ind][1]
        if ind % log_every_n == 0:
            print("Generating scaffold %d/%d" % (ind, data_len))
        scaffold = _generate_scaffold(smiles)
        if scaffold not in scaffolds:
            scaffolds[scaffold] = [ind]
        else:
            scaffolds[scaffold].append(ind)

    # Sort from largest to smallest scaffold sets
    scaffolds = {key: sorted(value) for key, value in scaffolds.items()}
    scaffold_sets = [
        scaffold_set for (scaffold, scaffold_set) in sorted(
            scaffolds.items(), key=lambda x: (len(x[1]), x[1][0]), reverse=True)
    ]
    return scaffold_sets


def scaffold_split(dataset, valid_size, test_size, seed=None, log_every_n=1000):
    train_size = 1.0 - valid_size - test_size
    scaffold_sets = generate_scaffolds(dataset)

    train_cutoff = train_size * len(dataset)
    valid_cutoff = (train_size + valid_size) * len(dataset)
    train_inds: list[int] = []
    valid_inds: list[int] = []
    test_inds: list[int] = []

    print("About to sort in scaffold sets")
    for scaffold_set in scaffold_sets:
        if len(train_inds) + len(scaffold_set) > train_cutoff:
            if len(train_inds) + len(valid_inds) + len(scaffold_set) > valid_cutoff:
                test_inds += scaffold_set
            else:
                valid_inds += scaffold_set
        else:
            train_inds += scaffold_set
    return train_inds, valid_inds, test_inds


def create_dataset(ids_path,ipc_path,sfs_path,mlg_path,frac,mode='pretrain'):
    ipc_mapper = ipcMapper(ipc_path,max_len=130)
    sfs_mapper = sfsMapper(sfs_path,max_len=75)
    mlg_mapper = mlgMapper(mlg_path)
    if mode == 'pretrain':
        dataset = SIGDataset(ids_path,ipc_mapper, sfs_mapper, mlg_mapper,True)
        length=len(dataset)
        train_size,test_size=(length-int(frac*length)),int(frac*length)
        train_dataset,test_dataset=torch.utils.data.random_split(dataset,[train_size,test_size])

        return train_dataset,test_dataset

    elif mode == 'finetune':
        dataset = FinetuneDataset(ids_path,mlg_path,ipc_mapper, sfs_mapper, mlg_mapper,True)
        length=len(dataset)
        train_size,val_size,test_size=(length-2*int(frac*length)),int(frac*length),int(frac*length)
        train_dataset,val_dataset,test_dataset=torch.utils.data.random_split(dataset,[train_size,val_size,test_size])        

        # train_idx, valid_idx, test_idx = scaffold_split(dataset, frac, frac)
        # train_dataset = Subset(dataset, train_idx)
        # val_dataset = Subset(dataset, valid_idx)
        # test_dataset = Subset(dataset, test_idx)
        # return train_dataset,val_dataset,test_dataset
    
        return train_dataset,val_dataset,test_dataset

def create_dataloader(dataset,batch_size,mode='pretrain',task_type='classification'):
    drop = False
    if mode == 'pretrain':
        collate = SIG_collate
        drop = True
    elif mode == 'finetune' and task_type == 'classification':
        collate = Finetune_collate_C
    elif mode == 'finetune' and task_type == 'regression':
        collate = Finetune_collate_R
    data_loader = DataLoader(
        dataset,
        batch_size=batch_size,
        shuffle=False,
        collate_fn= collate,
        drop_last=drop
    )
    return data_loader






import torch
from torch_geometric.data import InMemoryDataset


class PubChemDataset(InMemoryDataset):
    def __init__(self, path):
        super(PubChemDataset, self).__init__()
        self.data, self.slices = torch.load(path)
    
    def __getitem__(self, idx):
        return self.get(idx)


data = PubChemDataset('/home/hukun/code/Tmm-llama/data/PubChem324kV2/PubChem324kV2/pretrain.pt')
print(data[0])


import pubchempy
smi = data[0]["smiles"]
compounds = pubchempy.get_compounds(smi, namespace='smiles')
match = compounds[0]
iupac_name = match.iupac_name
print(iupac_name)




import torch
import pubchempy
import pandas as pd
import time
from torch_geometric.data import InMemoryDataset
import selfies as sf
from tqdm import tqdm


class PubChemDataset(InMemoryDataset):
    def __init__(self, path):
        super(PubChemDataset, self).__init__()
        self.data, self.slices = torch.load(path)
    
    def __getitem__(self, idx):
        return self.get(idx)

# 加载数据集
data = PubChemDataset('/home/hukun/code/Tmm-llama/data/PubChem324kV2/PubChem324kV2/pretrain.pt')

# 函数：获取iUPAC名称，添加超时处理
def get_iupac_name(smiles, timeout=60):
    try:
        compounds = pubchempy.get_compounds(smiles, namespace='smiles', timeout=timeout)
        if compounds:
            return compounds[0].iupac_name
        return None
    except Exception as e:
        print(f"Error fetching IUPAC name for {smiles}: {e}")
        return None

# 准备存储数据的列表
data_list = []
unfinished_indexes = []  # 用于记录未完成的索引

# 打开一个 CSV 文件以便实时保存
csv_file = './data/pubchem324k/pubchem_data.csv'
df_columns = ["CID", "SMILES", "text", "iupac name","SELFIES"]
df = pd.DataFrame(columns=df_columns)
df.to_csv(csv_file, index=False)  # 写入表头

# 遍历数据集
for i in tqdm(range(len(data)), desc="Processing", unit="sample"):
    sample = data[i]
    smiles = sample["smiles"]
    cid = sample["cid"]
    text = sample["text"]
    
    # 获取 iupac name
    iupac_name = get_iupac_name(smiles)
    selfies_ = sf.encoder(smiles)
    if iupac_name is not None:
        # 保存有效数据
        data_list.append([cid, smiles, text, iupac_name,selfies_])
        # 实时保存到 CSV 文件
        temp_df = pd.DataFrame([[cid, smiles, text, iupac_name,selfies_]], columns=df_columns)
        temp_df.to_csv(csv_file, mode='a', header=False, index=False)
    else:
        # 记录未完成的索引
        unfinished_indexes.append(i)


# 保存未完成数据的索引到一个文件
unfinished_file = './data/pubchem324k/unfinished_indexes.txt'
with open(unfinished_file, 'w') as f:
    for idx in unfinished_indexes:
        f.write(f"{idx}\n")

print("数据已成功保存为 CSV 文件！")
print(f"未完成的数据索引已保存到 {unfinished_file}！")



import argparse
import os
import ruamel.yaml as yaml
import numpy as np
import random
import time
import datetime
import json
from pathlib import Path
import torch
import torch.nn as nn
import torch.nn.functional as F
import utils
from model.TMMF import TMMF
from utils.utils import MetricLogger, SmoothedValue,AttrDict
from dataset import create_dataset, create_dataloader
from scheduler import create_scheduler
from optim import create_optimizer


def train(model, data_loader, optimizer, epoch, warmup_steps, device, scheduler, config):

    model.train()
    Total_params = 0
    Trainable_params = 0
    NonTrainable_params = 0
    for param in model.parameters():
        mulValue = np.prod(param.size())  # 
        Total_params += mulValue  
        if param.requires_grad:
            Trainable_params += mulValue  # trainable
        else:
            NonTrainable_params += mulValue  


    print(f'Total params: {Total_params}')
    print(f'Trainable params: {Trainable_params}')
    print(f'Non-trainable params: {NonTrainable_params}')
    metric_logger = MetricLogger(delimiter="  ")
    metric_logger.add_meter('lr', SmoothedValue(window_size=50, fmt='{value:.6f}'))
    metric_logger.add_meter('loss_msm', SmoothedValue(window_size=50, fmt='{value:.4f}'))
    metric_logger.add_meter('loss_sima', SmoothedValue(window_size=50, fmt='{value:.4f}'))
    metric_logger.add_meter('loss_simm', SmoothedValue(window_size=50, fmt='{value:.4f}'))
    
    header = 'Train Epoch: [{}]'.format(epoch)
    print_freq = 100   
    step_size = 100
    warmup_iterations = warmup_steps*step_size  


    for i, batch in enumerate(metric_logger.log_every(data_loader, print_freq, header)):
        
        optimizer.zero_grad()
        selfies,iupac,mol = batch['sfs_tokens']['selfies_tokens'],batch['ipc_tokens']['iupac_tokens'],batch['mlg']
        selfies = selfies.to(device)
        iupac = iupac.to(device)
        mol = mol.to(device)  

        _,loss_sima,loss_simm,loss_msm = model(mol,iupac,selfies)
        loss = loss_sima+loss_simm+loss_msm
        loss.backward()
        optimizer.step()    
        
        metric_logger.update(loss_msm=loss_msm.item())
        metric_logger.update(loss_sima=loss_sima.item())
        metric_logger.update(loss_simm=loss_simm.item())
        metric_logger.update(lr=optimizer.param_groups[0]["lr"])         
        
        if epoch==0 and i%step_size==0 and i<=warmup_iterations: 
            scheduler.step(i//step_size)        


    print("Averaged stats:", metric_logger.global_avg()) 
    return {k: "{:.3f}".format(meter.global_avg) for k, meter in metric_logger.meters.items()}    


def test(model,data_loader,device):
    i = 0
    num_total = 0
    correct_total = 0
    print_freq = 50
    model.eval()
    with torch.no_grad():
        for batch in data_loader:
            selfies,iupac,mol = batch['sfs_tokens']['selfies_tokens'],batch['ipc_tokens']['iupac_tokens'],batch['mlg']
            selfies = selfies.to(device)
            iupac = iupac.to(device)
            mol = mol.to(device)  
            correct,num = model.test(selfies,iupac,mol)
            correct_total += correct
            num_total += num
            i += 1
        
        if i%print_freq==0:
            print("iter:",i,"acc:",correct/num)
        
    print("total acc:",correct_total/num_total)


def main(args,config):

    device = torch.device(args.device)
    seed = args.seed
    torch.manual_seed(seed)
    np.random.seed(seed)

    max_epoch = config['schedular']['epochs']
    warmup_steps = config['schedular']['warmup_epochs']   

    print("creating dataset")
    train_dataset,test_dataset = create_dataset(
    config['dataset']['ids_path'],
    config['dataset']['ipc_path'],
    config['dataset']['sfs_path'],
    config['dataset']['mlg_path'],
    config['dataset']['frac'])

    train_loader = create_dataloader(train_dataset,config['dataset']['batch_size'])
    test_loader = create_dataloader(test_dataset,config['dataset']['batch_size'])

    print("creating model")
    model = TMMF(config)
    model = model.to(device)

    # 注释下面三行代码，可以从头开始训练
    # path = './pre_train/100k_checkpoint_09.pth'
    # checkpoint = torch.load(path)
    # model.load_state_dict(checkpoint['model'])

    arg_opt = AttrDict(config['optimizer'])
    optimizer = create_optimizer(arg_opt, model)
    arg_sche = AttrDict(config['schedular'])
    lr_scheduler, _ = create_scheduler(arg_sche, optimizer)  

    print("Start training")
    start_time = time.time()
    start_epoch = 0

    for epoch in range(start_epoch, max_epoch):
        
        if epoch>0:
            lr_scheduler.step(epoch+warmup_steps)  
        
        train_stats = train(model, train_loader, optimizer, epoch, warmup_steps, device, lr_scheduler, config)
        
        log_stats = {**{f'train_{k}': v for k, v in train_stats.items()},
                        'epoch': epoch,
                    }                     
        save_obj = {
            'model': model.state_dict(),
            'optimizer': optimizer.state_dict(),
            'lr_scheduler': lr_scheduler.state_dict(),
            'config': config,
            'epoch': epoch,
        }
        if (epoch+1)%5 == 0 or epoch==max_epoch-1:
            torch.save(save_obj, os.path.join(args.output_dir, '100k_checkpoint_test%02d.pth'%epoch))  
        
        with open(os.path.join(args.output_dir, "log.txt"),"a") as f:
            f.write(json.dumps(log_stats) + "\n")

    total_time = time.time() - start_time
    total_time_str = str(datetime.timedelta(seconds=int(total_time)))
    print('Training time {}'.format(total_time_str)) 

    print("Start testing")
    test(model,test_loader,device)



if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--config', default='./config/pre_train.yaml')
    parser.add_argument('--output_dir', default='pre_train/')
    parser.add_argument('--device', default=torch.device("cuda:0"))
    parser.add_argument('--seed', default=42, type=int)
    args = parser.parse_args()


    config = yaml.load(open(args.config, 'r'), Loader=yaml.Loader)

    Path(args.output_dir).mkdir(parents=True, exist_ok=True)
     
    main(args, config)